#include "shader.hpp"

static u32
compile_shader(std::string path, u32 shader_type)
{
    u32 id;
    std::string source = read_file(path);
    id = glCreateShader(shader_type);

    const char *p = source.c_str();
    glShaderSource(id, 1, &p, NULL);
    glCompileShader(id);

    /* error handling */
    int success;
    glGetShaderiv(id, GL_COMPILE_STATUS, &success);
    if (!success) {
        char errmsg[512];
        glGetShaderInfoLog(id, 512, NULL, errmsg);
        fprintf(stderr, "Shader compilation error: %s\n", errmsg);
    }

    return id;
}

static u32
create_program(VertexShader v, FragmentShader f)
{
    u32 id = glCreateProgram();

    glAttachShader(id, v.id);
    glAttachShader(id, f.id);
    glLinkProgram(id);

    /* error handling */
    int success;
    glGetProgramiv(id, GL_LINK_STATUS, &success);
    if (!success) {
        char errmsg[512];
        glGetProgramInfoLog(id, 512, NULL, errmsg);
        fprintf(stderr, "Shader linking error: %s\n", errmsg);
        glDeleteProgram(id);

        return 0;
    }

    return id;
}

VertexShader::VertexShader(std::string path)
{
    id = compile_shader(path, GL_VERTEX_SHADER);
}

VertexShader::~VertexShader()
{
    glDeleteShader(id);
}

FragmentShader::FragmentShader(std::string path)
{
    id = compile_shader(path, GL_FRAGMENT_SHADER);
}

FragmentShader::~FragmentShader()
{
    glDeleteShader(id);
}

Shader::Shader(VertexShader v, FragmentShader f)
{
    id = create_program(v, f);
}

Shader::Shader(std::string vpath, std::string fpath)
{
    id = create_program(VertexShader(vpath), FragmentShader(fpath));
}

void Shader::use() const { glUseProgram(id); }

void
Shader::set_int(const char *name, int value) const
{
    use();
    glUniform1i(glGetUniformLocation(id, name), value);
}

void
Shader::set_int(const std::string &name, int value) const
{
    set_int(name.c_str(), value);
}

void
Shader::set_float(const char *name, float value) const
{
    use();
    glUniform1f(glGetUniformLocation(id, name), value);
}

void
Shader::set_float(const std::string &name, float value) const
{
    set_float(name.c_str(), value);
}

void
Shader::set_mat4(const char *name, const glm::mat4 &mat) const
{
    use();
    glUniformMatrix4fv(glGetUniformLocation(id, name),
                       1,
                       GL_FALSE,
                       glm::value_ptr(mat));
}

void
Shader::set_mat4(const std::string &name, const glm::mat4 &mat) const
{
    set_mat4(name.c_str(), mat);
}
